import logging
from datetime import datetime
from pathlib import Path
from typing import Union

import pandas as pd

from .data import DATETIME_FORMAT, FORECASTED_COL, RUNTIME_COL

logger = logging.getLogger(__name__)


def _parse_datetime_cols(df: pd.DataFrame) -> pd.DataFrame:
    """Finds datetime columns in the DataFrame and converts them to datetime

    Args:
        df: pandas.DataFrame
    Returns:
        :class:`pandas.DataFrame` with datetime columns converted according to standard
        AEMO
        format
    """
    dt_cols = {
        "DATETIME",
        "EFFECTIVEDATE",
        "INTERVAL_DATETIME",
        "RUN_DATETIME",
        "AUTHORISEDDATE",
        "LASTCHANGED",
        "VERSION_DATETIME",
        "DAY",
        "PUBLISH_DATETIME",
        "LATEST_OFFER_DATETIME",
        "STARTDATE",
        "ENDDATE",
        "PERIOD_ENDING",
        "GENCONID_EFFECTIVEDATE",
        "BIDSETTLEMENTDATE",
        "SETTLEMENTDATE",
        "OFFERDATE",
    }
    dt_cols_present = dt_cols.intersection(set(df.columns.tolist()))
    for col in dt_cols_present:
        df.loc[:, col] = pd.to_datetime(df[col], format=DATETIME_FORMAT)
    return df


def _parse_id_cols(df: pd.DataFrame) -> pd.DataFrame:
    """Finds relevant ID columns in the DataFrame and converts them to categories

    Args:
        df: pandas.DataFrame
    Returns:
        :class:`pandas.DataFrame` with ID columns converted to categories
    """
    id_cols = {
        "CONSTRAINTID",
        "INTERCONNECTORID",
        "DUID",
        "CONNECTIONPOINTID",
        "PARTICIPANTID",
        "EXPORTGENCONID",
        "IMPORTGENCONID",
        "REGIONID",
    }
    id_cols_present = id_cols.intersection(set(df.columns.tolist()))
    for col in id_cols_present:
        df.loc[:, col] = df[col].astype("category")
    return df


def _parse_predispatch_seq_no(df: pd.DataFrame) -> pd.DataFrame:
    """Parses `PREDISPATCHSEQNO` as datetime and adds `PREDISPATCH_RUN_DATETIME`

    Args:
        df: pandas.DataFrame
    Returns:
        :class:`pandas.DataFrame` with additional column `PREDISPATCH_RUN_DATETIME`
    """
    df["PREDISPATCHSEQNO"] = df["PREDISPATCHSEQNO"].astype(int).astype(str)
    parsed = df["PREDISPATCHSEQNO"].str.extract(r"^([0-9]{8})([0-9]{2})$")
    year_month_day = pd.to_datetime(parsed[0], format="%Y%m%d")
    hour_min = ((parsed[1].astype(int) - 1) * pd.Timedelta(minutes=30)).add(
        pd.Timedelta(hours=4, minutes=30)
    )
    df["PREDISPATCH_RUN_DATETIME"] = year_month_day + hour_min  # type: ignore
    return df


def clean_forecast_csv(filepath_or_buffer: Union[str, Path]) -> pd.DataFrame:
    """Given a forecast csv filepath or buffer, reads and cleans the forecast csv.

    Cleans artefacts in the forecast csv files, including AEMO metadata at start of file
    and end of report line. Also removes any duplicate rows.

    Args:
        filepath_or_buffer: As for :func:`pandas.read_csv`
    Returns:
        Cleaned :class:`pandas.DataFrame` with forecast data
    Warning:
        Removes duplicate rows. Raises a warning when doing so.
    """
    # skip AEMO metadata
    df = pd.read_csv(filepath_or_buffer, skiprows=1, low_memory=False)
    # remove end of report line
    df = df.iloc[0:-1, :]
    drop_cols = df.columns.tolist()[0:4]
    df = df.drop(drop_cols, axis="columns")
    df = _parse_datetime_cols(df)
    df = _parse_id_cols(df)
    if "PREDISPATCHSEQNO" in df.columns:
        df = _parse_predispatch_seq_no(df)
    if any(dup_df := df.duplicated()):
        dup_rows = dup_df.loc[dup_df is True]
        logging.warning(
            "Duplicate rows detected. Dropping the following rows:\n" + f"{dup_rows}"
        )
        df = df.drop_duplicates()
    return df


def _filter_on_datetime_col(
    df: pd.DataFrame, dt_col: str, start: datetime, end: datetime
) -> pd.DataFrame:
    """Filter the given datetime column based on the supplied start and end times

    Args:
        df: pandas.DataFrame
        dt_col: Datetime columns in :attr:`df`
        start: Start datetime
        end: End datetime
    Returns
        DataFrame with datetime filtering applied on :attr:`dt_col`.
    """
    df = df.loc[df[dt_col] >= start, :]
    df = df.loc[df[dt_col] <= end, :]
    return df


def apply_run_and_forecasted_time_filters(
    df: pd.DataFrame,
    run_start: datetime,
    run_end: datetime,
    forecasted_start: datetime,
    forecasted_end: datetime,
    forecast_type: str,
) -> pd.DataFrame:
    """Applies filtering for run times (i.e. :term:`run_start` and :term:`run_end`) and
    forecasted times (i.e. :term:`forecasted_start` and :term:`forecasted_end`).

    Datetime filtering is applied to a column fetched from lookup tables that map
    the relevant column name to each :term:`forecast type`. If the run time/forecasted
    column obtained from the lookup is not present in the DataFrame, the respective
    filter is not applied.

    Args:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
        forecast_type: One of :data:`nemseer.forecast_types`.
    Returns:
        DataFrame with appropriate datetime filtering applied.
    """
    (runtime_col, forecasted_col) = (
        RUNTIME_COL[forecast_type],
        FORECASTED_COL[forecast_type],
    )
    for (start, end), col in zip(
        ((run_start, run_end), (forecasted_start, forecasted_end)),
        (runtime_col, forecasted_col),
    ):
        if col in df.columns:
            df = _filter_on_datetime_col(df, col, start, end)
    return df



import logging
import shutil
from datetime import datetime
from itertools import cycle
from pathlib import Path
from re import match
from typing import Dict, Generator, List
from zipfile import BadZipFile, ZipFile

import psutil
import requests
from attrs import define, field
from bs4 import BeautifulSoup
from tqdm.auto import tqdm

from .data import (
    ENUMERATED_TABLES,
    FORECAST_TYPES,
    INVALID_STUBS_FILE,
    MMSDM_ARCHIVE_URL,
    PREDISP_ALL_DATA,
    USER_AGENTS,
)
from .data_handlers import clean_forecast_csv
from .query import (
    Query,
    _construct_sqlloader_filename,
    _enumerate_tables,
    generate_sqlloader_filenames,
)

logger = logging.getLogger(__name__)


def _validate_forecast_type(forecast_type: str):
    """Check user-supplied forecast type is valid"""
    if forecast_type not in FORECAST_TYPES:
        raise ValueError(f"Forecast type should be one of {FORECAST_TYPES}")


def _build_useragent_generator(n: int) -> Generator[str, None, None]:
    """Generator function that cycles through user agents for GET requests.

    Generator function that cycles through user agents to yield n user agents
    in total. Doing so minimises 403 Forbidden errors when scraping.

    Args:
        n: Number of user agents, i.e. number of GET requests.
    Yields:
        useragent: A user agent.
    """
    inf_agents = cycle(USER_AGENTS)
    n_iterator = range(n)
    for _, useragent in zip(n_iterator, inf_agents):
        yield useragent


def _build_nemweb_get_header(useragent: str) -> Dict[str, str]:
    """Builds request header for GET requests from NEMWeb

    Args:
        useragent: User-Agent string to use
    Returns:
        Dict that can be used as a request header
    """
    header = {
        "Host": "www.nemweb.com.au",
        "User-Agent": useragent,
        "Accept": (
            "text/html,application/xhtml+xml,application/xml;"
            + "q=0.9,image/avif,image/webp,*/*;q=0.8"
        ),
        "Accept-Language": "en-US,en;q=0.5",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
        "Upgrade-Insecure-Requests": "1",
    }
    return header


def _request_content(
    url: str, useragent: str, additional_header: Dict = {}
) -> requests.Response:
    """Initiates a GET request with header information.

    Args:
        url: URL for GET request.
        useragent: User-Agent to use in header.
        additional_header: Empty dictionary as default. Can be used to add
            additional header information to GET request.
    Returns:
        requests Response object.
    """
    header = _build_nemweb_get_header(useragent)
    if additional_header:
        header.update(additional_header)
    r = requests.get(url, headers=header)
    return r


def _rerequest_to_obtain_soup(
    url: str, useragent: str, additional_header: Dict = {}
) -> BeautifulSoup:
    """Continually launches requests until a 200 (OK) code is returned.

    Args:
        url: URL for GET request.
        useragent: User-Agent to use in header.
        additional_header: Empty dictionary as default. Can be used to add
            additional header information to GET request.

    Returns:
        BeautifulSoup object with parsed HTML.

    """
    r = _request_content(url, useragent, additional_header=additional_header)
    while (ok := r.status_code == requests.status_codes.codes["OK"]) < 1:
        r = _request_content(url, useragent, additional_header=additional_header)
        if r.status_code == requests.status_codes.codes["OK"]:
            ok += 1
    soup = BeautifulSoup(r.content, "html.parser")
    return soup


def _construct_yearmonth_url(
    year: int, month: int, forecast_type: str, all_data: bool = False
) -> str:
    """Constructs MMSDM Historical Data SQLLoader Base URL for a given year and month.

    Handles exceptions to naming rules and complete tables (`PREDISP_ALL_DATA`)
    for `PREDISPATCH` when `forecast_type` = "PREDISPATCH" and `all_data` = True.

    N.B. Can be extended to handle `P5MIN_ALL_DATA`, but this appears to be similar to
    `P5MIN` data in `DATA` directory. Files in this directory have `ALL` in the name.

    Args:
        year: Year
        month: Month
        forecast_type: One of :data:`nemseer.forecast_types`
        all_data (optional): Default False. Points to `ALL_DATA` folder for PD
    Returns:
        Constructed URL as a string.
    """
    base_url = (
        MMSDM_ARCHIVE_URL
        + f"{year}/MMSDM_{year}_"
        + f'{str(month).rjust(2, "0")}/MMSDM_Historical_Data_SQLLoader/'
    )
    if forecast_type == "PREDISPATCH" and all_data:
        data_url = base_url + "PREDISP_ALL_DATA/"
    else:
        data_url = base_url + "DATA/"
    return data_url


def _construct_sqlloader_forecastdata_url(
    year: int, month: int, forecast_type: str, table: str
) -> str:
    """Constructs URL that points to a MMSDM Historical Data SQLLoader zip file

    Handles exceptions to naming rules and complete tables (`PREDISP_ALL_DATA`)
    for `PREDISPATCH`.

    Args:
        year: Year
        month: Month
        forecast_type: One of :data:`nemseer.forecast_types`
    Returns:
        URL to zip file
    """
    if (
        forecast_type == "PREDISPATCH"
        and (table_basename := match(r"([A-Z_]*)[0-9]?", table))
        and table_basename.group(1) in PREDISP_ALL_DATA
    ):
        data_url = _construct_yearmonth_url(year, month, forecast_type, all_data=True)
    else:
        data_url = _construct_yearmonth_url(year, month, forecast_type)
    fn = _construct_sqlloader_filename(year, month, forecast_type, table)
    url = data_url + fn + ".zip"
    return url


def _get_captured_group_from_links(url: str, regex: str) -> List[str]:
    """Returns list of unique captured groups from MMSDM Historical Data SQLLoader page

    For a year and month in the MMSDM Historical Data SQLLoader, returns captured groups
    associated with a particular `forecast_type`. Primarily used to obtain table names.

    Args:
        year: Year
        month: Month
        forecast_type: One of :data:`nemseer.forecast_types`
        regex: Regular expression pattern, with one group capture
    Returns:
        A list of unique captured groups (one for each link on the page of tables)
    """
    soup = _rerequest_to_obtain_soup(url, next(_build_useragent_generator(1)))
    links = [link.get("href") for link in soup.find_all("a")]
    tables = []
    for link in links:
        if mo := match(regex, link):
            tables.append(mo.group(1).lstrip("_"))
    return list(set(tables))


def get_sqlloader_forecast_tables(
    year: int, month: int, forecast_type: str, actual: bool = False
) -> List[str]:
    """Requestable tables of particular forecast type on MMSDM Historical Data SQLLoader

    If :attr:`actual` = False, provides a list of tables that can be requested via
    `nemseer`.

    If :attr:`actual` = True, returns actual tables available via NEMWeb, including all
    tables that are enumerated.

    N.B.:
      - Removes numbering from enumerated tables for `P5MIN`
        - e.g. `CONSTRAINTSOLUTION(x)` are all reduced to `CONSTRAINTSOLUTION`

    Args:
        year: Year
        month: Month
        forecast_type: One of :data:`nemseer.forecast_types`

    Returns:
        List of tables associated with that forecast type for that period
    """
    _validate_forecast_type(forecast_type)
    if actual:
        table_capture = f".*/PUBLIC_DVD_{forecast_type}([A-Z_0-9]*)_[0-9]*.zip"
    else:
        table_capture = f".*/PUBLIC_DVD_{forecast_type}([A-Z_]*)[0-9]?_[0-9]*.zip"
    data_url = _construct_yearmonth_url(year, month, forecast_type)
    tables = _get_captured_group_from_links(data_url, table_capture)
    if forecast_type == "PREDISPATCH":
        predisp_all_url = _construct_yearmonth_url(
            year, month, forecast_type, all_data=True
        )
        predisp_all_tables = _get_captured_group_from_links(
            predisp_all_url, table_capture
        )
        tables.extend(predisp_all_tables)
    return sorted(tables)


def get_sqlloader_years_and_months() -> Dict[int, List[int]]:
    """Years and months with data on NEMWeb MMSDM Historical Data SQLLoader

    Returns:
        Months mapped to each year. Data is available for each of these months.
    """

    def _get_months(url: str, useragent: str) -> List[int]:
        """Pull months from scraped links with YYYY-MM date format

        Args:
            url: url for GET request.
            header: useragent to pass to GET request.
        Returns:
            List of unique months (as integers).
        """
        referer_header = {"Referer": MMSDM_ARCHIVE_URL}
        soup = _rerequest_to_obtain_soup(
            url, useragent, additional_header=referer_header
        )
        months = []
        for link in soup.find_all("a"):
            url = link.get("href")
            findmonth = match(r".*[0-9]{4}_([0-9]{2})", url)
            if not findmonth:
                continue
            else:
                month = findmonth.group(1)
                months.append(int(month))
        unique = list(set(months))
        return unique

    useragent = next(_build_useragent_generator(1))
    soup = _rerequest_to_obtain_soup(MMSDM_ARCHIVE_URL, useragent)
    links = soup.find_all("a")
    nlinks = len(links)
    yearmonths = {}
    for useragent, link in zip(_build_useragent_generator(nlinks), links):
        url = link.get("href")
        findyear = match(r".*([0-9]{4}).*", url)
        if not findyear:
            continue
        else:
            year = int(findyear.group(1))
            months = _get_months(MMSDM_ARCHIVE_URL + f"{year}/", useragent)
            yearmonths[year] = months
    return yearmonths


def get_unzipped_csv(url: str, raw_cache: Path) -> None:
    """Unzipped (single) csv file downloaded from `url` to :term:`raw_cache`

    This function:

    1. Downloads zip file in chunks to limit memory use and enable progress bar
    2. Validates that the zip contains a single file that has the same name as the zip
    3. If the zip file is invalid, writes the file stub to `.invalid_aemo_files.txt`

    Args:
        url: URL of zip
        raw_cache: Path to save zip. See :term:`raw_cache`.
    Returns:
        None. Extracts csvs to :attr:`raw_cache`.
    """

    def _invalid_zip_to_file(invalid_files: Path, filename: str) -> None:
        """Ensure that any invalid file is noted in the `invalid_files` text file"""
        with open(invalid_files, "a+") as f:
            f.seek(0)
            existing = [line.strip() for line in f.readlines()]
            if filename in existing:
                pass
            else:
                f.write(f"{filename}\n")
        return None

    file_name = Path(url).name
    header = _build_nemweb_get_header(next(_build_useragent_generator(1)))
    file_path = raw_cache / Path(file_name)
    with requests.get(url, headers=header, stream=True) as resp:
        total_length = int(resp.headers.get("Content-Length", 0))
        resp.raise_for_status()
        with tqdm.wrapattr(resp.raw, "read", desc=file_name, total=total_length) as raw:
            with open(file_path, "wb") as fout:
                shutil.copyfileobj(raw, fout)
    z = ZipFile(file_path)
    if (
        len(csvfn := z.namelist()) == 1
        and (zfn := match(".*DATA/(.*).zip", url))
        and (fn := match("(.*).[cC][sS][vV]", csvfn.pop()))
        and (fn.group(1) == zfn.group(1))
    ):
        try:
            z.extractall(raw_cache)
            z.close()
        except BadZipFile:
            logging.error(f"{z.testzip()} invalid or corrupted")
            invalid_files = raw_cache / Path(INVALID_STUBS_FILE)
            _invalid_zip_to_file(invalid_files, fn.group(1))
        Path(file_path).unlink()
    else:
        raise ValueError(f"Unexpected contents in zipfile from {url}")


def _validate_tables_on_run_start(instance, attribute, value) -> None:
    """Validates tables for the provided forecast type.

    Checks user-supplied tables against tables available in MMS Historical
    Data SQLLoader for the month and year of run_start.
    """
    start_dt = instance.run_start
    actual_tables = get_sqlloader_forecast_tables(
        start_dt.year, start_dt.month, instance.forecast_type, actual=True
    )
    requestable_tables = get_sqlloader_forecast_tables(
        start_dt.year, start_dt.month, instance.forecast_type, actual=False
    )
    if not set(value).issubset(set(actual_tables)):
        raise ValueError(
            "Table(s) not available from MMS Historical Data SQL Loader"
            + f" (for {start_dt.month}/{start_dt.year}).\n"
            + f"Tables include: {requestable_tables}"
        )


@define(kw_only=True)
class ForecastTypeDownloader:
    """:class:`ForecastTypeDownloader` can initiate csv downloads and convert
    :term:`raw_cache` csvs to the parquet format.

    Attributes:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecast_type: One of :data:`nemseer.forecast_types`
        tables: Table or tables required. A single table can be supplied as
            a string. Multiple tables can be supplied as a list of strings.
        raw_cache: Path to download raw data to. Can reuse or build a
            new :term:`raw_cache`.
    """

    run_start: datetime
    run_end: datetime
    forecast_type: str
    tables: List[str] = field(validator=_validate_tables_on_run_start)
    raw_cache: Path

    @classmethod
    def from_Query(cls, query: Query) -> "ForecastTypeDownloader":
        """Constructor method for :class:`ForecastTypeDownloader` from
        :class:`Query <nemseer.query.Query>`"""
        tables = query.tables
        for ftype in ENUMERATED_TABLES:
            if query.forecast_type == ftype:
                for table, enumerate_to in ENUMERATED_TABLES[ftype]:
                    if table in tables:
                        tables = _enumerate_tables(tables, table, enumerate_to)

        return cls(
            run_start=query.run_start,
            run_end=query.run_end,
            forecast_type=query.forecast_type,
            tables=tables,
            raw_cache=query.raw_cache,
        )

    def download_csv(self) -> None:
        """Downloads and unzips zip files given query loaded into
        :class:`ForecastTypeDownloader`

        This method will only download and unzip the relevant zip/csv if the
        corresponding `.parquet` file is not located in the specified :attr:`raw_cache`.
        """
        filename_data = generate_sqlloader_filenames(
            self.run_start, self.run_end, self.forecast_type, self.tables
        )
        invalid_or_corrupted_stubfile = self.raw_cache / Path(INVALID_STUBS_FILE)
        for metadata in filename_data.keys():
            fname = filename_data[metadata]
            (year, month, table) = metadata
            if (self.raw_cache / Path(fname + ".parquet")).exists():
                logging.info(f"{table} for {month}/{year} in raw_cache")
                continue
            else:
                if invalid_or_corrupted_stubfile.exists():
                    with open(invalid_or_corrupted_stubfile, "r") as f:
                        invalid_or_corrupted = f.readlines()
                    check_files = [f.strip() for f in invalid_or_corrupted]
                    if fname in check_files:
                        logging.warning(
                            f"{fname} previously found to be invalid/corrupted. "
                            + "Skipping download for this file. "
                            + "If downloading manually works, remove from "
                            + ".invalid_aemo_files.txt in raw_cache. "
                            + "Otherwise, contact AEMO."
                        )
                        continue
                url = _construct_sqlloader_forecastdata_url(
                    year, month, self.forecast_type, table
                )
                logger.info(f"Downloading and unzipping {table} for {month}/{year}")
                get_unzipped_csv(url, self.raw_cache)

    def convert_to_parquet(self, keep_csv=False) -> None:
        """Converts all CSVs in the :attr:`raw_cache` to parquet

        Warning:
            A warning is printed if the filesize is greater than half of available
            memory as :class:`pandas.DataFrame` consumes more than the file size in
            memory.
        """
        csvs = list(Path(self.raw_cache).glob("*.[Cc][Ss][Vv]"))
        for csv in csvs:
            if csv.stat().st_size * 2 >= psutil.virtual_memory().available:
                logging.warning(
                    f"Attempting to convert {csv} to parquet,"
                    + " but your available system memory may be too low for this."
                )
            df = clean_forecast_csv(csv)
            parquet_name = csv.name[0:-3] + "parquet"
            if not csv.with_name(parquet_name).exists():
                logging.info(f"Converting {csv.name} to parquet")
                df.to_parquet(csv.with_name(parquet_name))
            else:
                logging.info(f"{parquet_name} already exists")
            if not keep_csv:
                csv.unlink()



import logging
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from attrs import converters, define, field, validators
from dateutil import rrule

from .data import DATETIME_FORMAT, ENUMERATED_TABLES, FORECAST_TYPES

logger = logging.getLogger(__name__)


def _dt_converter(value: str) -> datetime:
    """Convert string to datetime.

    Args:
        value: String with format %Y/%m/%d %H:%M
    Returns:
        Datetime object
    Raises:
        ValueError: If provided datetime string is invalid
    """
    try:
        return datetime.strptime(value, DATETIME_FORMAT)
    except ValueError:
        raise ValueError(
            "Datetime invalid. Datetime should be provided as follows: yyyy/mm/dd HH:MM"
        )


def _tablestr_converter(value: Union[str, List[str]]) -> List[str]:
    """Returns a list of table strings, even if a single string is provided

    Args:
        value: Table string or list of table strings
    Returns:
        List of strings
    """
    if type(value) is str:
        return [value]
    else:
        return list(value)


def _validate_run_chronology(instance, attribute, value):
    """Validates run_start against run_end"""
    if value > instance.run_end:
        raise ValueError(
            "Forecast end datetime must be greater than or equal to"
            + " run start datetime."
        )


def _validate_forecasted_chronology(instance, attribute, value):
    """Validated forecasted_start against forecasted_end"""
    if value > instance.forecasted_end:
        raise ValueError(
            "Forecasted end datetime must be greater than or equal to"
            + " forecasted start datetime."
        )


def _validate_relative_chronology(instance, attribute, value) -> None:
    """Validates run_start against forecasted_start"""
    if value > instance.forecasted_start:
        raise ValueError(
            "Forecasted start datetime should be at or after run start datetime."
        )


def _validate_path(instance, attribute, value) -> None:
    """Check the path is a directory and creates it if it is not"""
    if not value.is_dir():
        value.mkdir()
        logger.info(f"Created directory at {value.absolute()}")


def _validate_raw_not_processed(instance, attribute, value) -> None:
    """Check that :attr:`raw_cache` and :attr:`processed_cache` are distinct."""
    if instance.processed_cache:
        if value.absolute() == instance.processed_cache.absolute():
            raise ValueError(
                f"{attribute.name} should be distinct from processed_cache"
            )


def _enumerate_tables(tables: List[str], table_str: str, range_to: int) -> List[str]:
    """Given a table name, populates a list with enumerated table names

    For example, given 'CONSTRAINTSOLUTION' and `range_to`=3, will populate
    `tables` with ['CONSTRAINTSOLUTION1',...,'CONSTRAINTSOLUTION3'].

    Args:
        tables: Table list
        table_str: Table string to enumerate
        range_to: Integer to enumerate to
    Returns:
        `tables` with enumerated `table_str`
    """
    tables.remove(table_str)
    for i in range(1, range_to + 1):
        tables.append(f"{table_str}{i}")
    return tables


def _construct_sqlloader_filename(
    year: int, month: int, forecast_type: str, table: str
) -> str:
    """Constructs filename without file type

    Args:
        year: Year
        month: Month
        forecast_type: One of :data:`nemseer.forecast_types`. See :term:`forecast types`
        table: The name of the table required
    Returns:
        Filename string without file type
    """
    (stryear, strmonth) = (str(year), str(month).rjust(2, "0"))
    if forecast_type == "PREDISPATCH" and table != "MNSPBIDTRK":
        prefix = f"PUBLIC_DVD_{forecast_type}{table}"
    else:
        prefix = f"PUBLIC_DVD_{forecast_type}_{table}"
    fn = prefix + f"_{stryear}{strmonth}010000"
    return fn


def generate_sqlloader_filenames(
    run_start: datetime,
    run_end: datetime,
    forecast_type: str,
    tables: List[str],
) -> Dict[Tuple[int, int, str], str]:
    """Generates MMSDM Historical Data SQLLoader file names based on provided query data

    Returns a tuple of query metadata (`table`, `year`, `month`) mapped to each filename

    Args:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecast_type: One of :data:`nemseer.forecast_types`.
        tables: Table or tables required, provided as a List.
    Returns:
        A tuple of query metadata (`table`, `year`, `month`) mapped to each
        format-agnostic (:term:`SQLLoader`) filename
    """
    intervening_dates = rrule.rrule(rrule.MONTHLY, dtstart=run_start, until=run_end)
    filename_data = {}
    for ftype in ENUMERATED_TABLES:
        if forecast_type == ftype:
            for table, enumerate_to in ENUMERATED_TABLES[ftype]:
                if table in tables:
                    tables = _enumerate_tables(tables, table, enumerate_to)
    for table in tables:
        for date in intervening_dates:
            (year, month) = (date.year, date.month)
            fname = _construct_sqlloader_filename(year, month, forecast_type, table)
            filename_data[(year, month, table)] = fname
    return filename_data


@define
class Query:
    """:class:`Query` validates user inputs and dispatches data downloaders and
    compilers

    Construct :class:`Query` using the :meth:`Query.initialise()` constructor. This
    ensures query metadata is constructed approriately.

    Query:

    - Validates user input data
        - Checks datetimes fit :attr:`yyyy/mm/dd HH:MM` format
        - Checks datetime chronology (e.g. end is after start)
        - Checks requested datetimes are valid for each :term:`forecast type`
        - Validates :term:`forecast type`
        - Validates user-requested tables against what is available on NEMWeb
    - Retains query metadata (via constructor class method
      :meth:`nemseer.query.Query.initialise`)
    - Can dispatch
      :class:`ForecastTypeDownloader <nemseer.downloader.ForecastTypeDownloader>` and
      :class:`DataCompiler <nemseer.data_compilers.DataCompiler>`

    Attributes:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
        forecast_type: One of :data:`nemseer.forecast_types`.
        tables: Table or tables required. A single table can be supplied as
            a string. Multiple tables can be supplied as a list of strings.
        metadata: Metadata dictionary. Constructed by :meth:`Query.initialise()`.
        raw_cache (optional): Path to build or reuse :term:`raw_cache`.
        processed_cache (optional): Path to build or reuse :term:`processed_cache`.
            Should be distinct from :attr:`raw_cache`

    """

    run_start: datetime = field(
        converter=_dt_converter,
        validator=[_validate_run_chronology, _validate_relative_chronology],
    )
    run_end: datetime = field(converter=_dt_converter)
    forecasted_start: datetime = field(
        converter=_dt_converter, validator=[_validate_forecasted_chronology]
    )
    forecasted_end: datetime = field(converter=_dt_converter)
    forecast_type: str = field(validator=validators.in_(FORECAST_TYPES))
    tables: List[str] = field(converter=_tablestr_converter)
    metadata: Dict
    raw_cache: Path = field(
        converter=Path,
        validator=[
            _validate_path,
            _validate_raw_not_processed,
        ],
    )
    processed_cache: Optional[Path] = field(
        default=None,
        converter=converters.optional(Path),
        validator=validators.optional(_validate_path),
    )

    @classmethod
    def initialise(
        cls,
        run_start: str,
        run_end: str,
        forecasted_start: str,
        forecasted_end: str,
        forecast_type: str,
        tables: Union[str, List[str]],
        raw_cache: str,
        processed_cache: Optional[str] = None,
    ) -> "Query":
        """Constructor method for :class:`Query`. Assembles query metatdata."""
        metadata = {
            "run_start": run_start,
            "run_end": run_end,
            "forecasted_start": forecasted_start,
            "forecasted_end": forecasted_end,
            "forecast_type": forecast_type,
            "tables": tables,
        }
        return cls(
            run_start=run_start,  # type: ignore
            run_end=run_end,  # type: ignore
            forecasted_start=forecasted_start,  # type: ignore
            forecasted_end=forecasted_end,  # type: ignore
            forecast_type=forecast_type,  # type: ignore
            tables=tables,  # type: ignore
            metadata=metadata,  # type: ignore
            raw_cache=raw_cache,  # type: ignore
            processed_cache=processed_cache,  # type: ignore
        )

    def check_data_in_cache(self) -> bool:
        """Checks whether *all* requested data is already in the :attr:`raw_cache` as
        parquet

        :meth:`nemseer.downloader.ForecastTypeDownloader.download_csv()`
        handles partial :attr:`raw_cache` completeness

        If all requested data is already in the :attr:`raw_cache` as parquet,
        returns True. Otherwise returns False.
        """
        fnames = generate_sqlloader_filenames(
            self.run_start, self.run_end, self.forecast_type, self.tables
        ).values()
        check = [
            (self.raw_cache / Path(fname + ".parquet")).exists() for fname in fnames
        ]
        if all(check):
            logging.info(f"Query raw data already downloaded to {self.raw_cache}")
            return True
        else:
            return False



from datetime import datetime, timedelta

from nemseer.data import DATETIME_FORMAT


def _determine_last_market_day_end_for_half_hourly(dt: datetime) -> datetime:
    """Returns end of last trading day for which price offer submission has closed by
    the supplied datetime.

    Only valid for forecasts datetimes with half-hourly increments, and that are run
    at or less frequently than every half hour.

    Args:
        dt: Datetime to find end of next
    Returns:
        Datetime corresponding to end of the last trading day for which price offer
        submission has closed by the supplied datetime.
    """
    if dt.hour >= 13:
        end_date = (dt + timedelta(days=2)).date()
    else:
        end_date = (dt + timedelta(days=1)).date()
    end_dt = datetime(
        year=end_date.year,
        month=end_date.month,
        day=end_date.day,
        hour=4,
        minute=0,
    )
    return end_dt


def validate_P5MIN_datetime_inputs(
    run_start: datetime,
    run_end: datetime,
    forecasted_start: datetime,
    forecasted_end: datetime,
) -> None:
    """Validates `P5MIN` forecast datetime inputs

    From `AEMO MMS Data Model Report <https://nemweb.com.au/Reports/Current/
    MMSDataModelReport/Electricity/MMS%20Data%20Model%20Report_files/MMS_230.htm#1>`_:

      The 5-minute Predispatch cycle runs every 5-minutes to produce a dispatch and
      pricing schedule to a 5-minute resolution covering the next hour,
      a total of twelve periods.

    Validation checks:

    Check 1:
      Minute component of datetime inputs is on a 5 minute basis
    Check 2:
      :attr:`forecasted_end` is not more than 55 minutes (12 cycles) from
      :attr:`run_end`

    These 12 dispatch cycles include the immediate interval
    (i.e. where `RUN_DATETIME` = `INTERVAL_DATETIME`)

    Args:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
    Raises:
        ValueError: If any validation conditions are failed.
    """
    # Check 1
    acceptable_minutes = set(range(0, 60, 5))
    for dt_input in (run_start, run_end, forecasted_start, forecasted_end):
        if dt_input.minute not in acceptable_minutes:
            raise ValueError(
                "P5MIN is run every 5 minutes.\n"
                + "Minutes in datetime inputs should correspond to: "
                + f"{acceptable_minutes}"
            )
    # Check 2
    if forecasted_end > (allowed := run_end + timedelta(minutes=55)):
        print_allowed = allowed.strftime(DATETIME_FORMAT)
        raise ValueError(
            "For P5MIN, forecasted_end must be within 55 minutes of run_end.\n"
            + f"This corresponds to {print_allowed} for the provided run_end"
        )
    return None


def validate_PREDISPATCH_datetime_inputs(
    run_start: datetime,
    run_end: datetime,
    forecasted_start: datetime,
    forecasted_end: datetime,
) -> None:
    """Validates `PREDISPATCH` forecast datetime inputs

    From `AEMO Pre-dispatch SOP <https://www.aemo.com.au/-/media/files/electricity/nem/
    security_and_reliability/power_system_ops/procedures/
    so_op_3704-predispatch.pdf?la=en>`_:

      Currently AEMO runs pre-dispatch every half hour, on the half hour for each
      30-minute period up to and including the last 30-minute period of the last trading
      day for which bid band prices have closed. As changes to bid band prices for the
      next trading day close at 1230 hours EST, AEMO will at 1230 hours,
      publish pre-dispatch for all 30-minute periods up to the end of the
      next trading day.

    Noting that:

    - A market/trading day extends from 0400 to 0400 on the next day.
    - Pre-dispatch executed at 1230 hours is associated with the 1300 hours run time.
      That is, `PREDISPATCHSEQ` corresponding to 13:00 contains bids for the
      next trading day.

    Validation checks:

    Check 1:
      Minute component of datetime inputs is on a 30 minute basis
    Check 2:
      :attr:`forecasted_end` is no later than the end of the last trading day for which
      bid band prices have closed (the end of that day being 04:00) by
      :attr:`run_end`

    Args:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
    Raises:
        ValueError: If any validation conditions are failed.
    """
    # Check 1
    acceptable_minutes = set((0, 30))
    for dt_input in (run_start, run_end, forecasted_start, forecasted_end):
        if dt_input.minute not in acceptable_minutes:
            raise ValueError(
                "PREDISPATCH/PDPASA is run every 30 minutes.\n"
                + "Minutes in datetime inputs should correspond to: "
                + f"{acceptable_minutes}"
            )
    # Check 2
    check_dt = _determine_last_market_day_end_for_half_hourly(run_end)
    if forecasted_end > check_dt:
        print_allowed = check_dt.strftime(DATETIME_FORMAT)
        raise ValueError(
            "For PREDISPATCH/PDPASA, forecasted_end must be no later than"
            + f" {print_allowed} based on the supplied run_end"
        )
    return None


def validate_PDPASA_datetime_inputs(
    run_start: datetime,
    run_end: datetime,
    forecasted_start: datetime,
    forecasted_end: datetime,
) -> None:
    """Validates `PDPASA` forecast datetime inputs

    Points to :func:`validate_PREDISPATCH_datetime_inputs()` as validation for
    PREDISPATCH and PDPASA are the same.

    Args:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
    Raises:
        ValueError: If any validation conditions are failed.
    """
    validate_PREDISPATCH_datetime_inputs(
        run_start, run_end, forecasted_start, forecasted_end
    )
    return None


def validate_STPASA_datetime_inputs(
    run_start: datetime,
    run_end: datetime,
    forecasted_start: datetime,
    forecasted_end: datetime,
) -> None:
    """Validates `STPASA` forecast datetime inputs

    From `AEMO PASA Outputs <https://aemo.com.au/energy-systems/electricity/
    national-electricity-market-nem/data-nem/market-management-system-mms-data/
    projected-assessment-of-system-adequacy-pasa>`_:

      [ST PASA] is published every 2 hours and provides detailed disclosure of
      short-term is published every 2 hours and provides detailed disclosure of
      short-term power-system supply/demand balance prospects for six days
      following the next trading day. The information is provided for each half-hour
      within the report period

    Noting that:

    - A market/trading day extends from 0400 to 0400 on the next day.
    - `ST PASA` is the "reverse" of `PREDISPATCH`
      - `ST PASA` **starts** after the end of the next trading day for which bids have
      been submitted

    The National Electricity Rules and some of AEMO's procedures state that ST PASA
    is run every two hours. The frequency was increased to hourly. See
    `Spot Market Operations Timetable <https://www.aemo.com.au/-/media/Files/
    Electricity/NEM/Security_and_Reliability/Dispatch/
    Spot-Market-Operations-Timetable.pdf>`_.


    Validation checks:

    Check 1:
      Minute component of forecast datetimes is on an hourly basis (i.e. 0 minutes)
    Check 2:
      Minute component of forecasted datetimes is on a 30 minute basis
    Check 3:
      :attr:`forecasted_start` is not equal to or earlier than the end of the
      last trading day for which bid band prices have closed
      (the end of that day being 04:00) by :attr:`run_start`
    Check 4:
      :attr:`forecasted_end` is no later than 6 days from the end of the last trading
      day for which bid band prices have closed by :attr:`run_end`

    Args:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
    Raises:
        ValueError: If any validation conditions are failed.
    """
    # Check 1
    for run_input in (run_start, run_end):
        if run_input.minute != 0:
            raise ValueError("ST PASA run_start and run_end must be on the hour")
    # Check 2
    forecasted_minutes = set((0, 30))
    for dt_input in (forecasted_start, forecasted_end):
        if dt_input.minute not in forecasted_minutes:
            raise ValueError(
                "ST PASA forecasts are provided for every half hour "
                + "in the forecast period\n"
                + " Minutes in forecasted_start and forecasted_end "
                + f" should correspond to: {forecasted_minutes}"
            )
    # Check 3
    end_of_last_for_start = _determine_last_market_day_end_for_half_hourly(run_start)
    start_check_dt = end_of_last_for_start + timedelta(minutes=30)
    if forecasted_start < start_check_dt:
        print_allowed = start_check_dt.strftime(DATETIME_FORMAT)
        raise ValueError(
            f"For ST PASA, forecasted_start must be no earlier than {print_allowed} "
            + "based on the supplied run_start"
        )
    # Check 4
    end_of_last_for_end = _determine_last_market_day_end_for_half_hourly(run_end)
    end_check_dt = end_of_last_for_end + timedelta(days=6)
    if forecasted_end > end_check_dt:
        print_allowed = end_check_dt.strftime(DATETIME_FORMAT)
        raise ValueError(
            f"For ST PASA, forecasted_end must be no later than {print_allowed} "
            + "based on the supplied run_end"
        )
    return None


def validate_MTPASA_datetime_inputs(
    run_start: datetime,
    run_end: datetime,
    forecasted_start: datetime,
    forecasted_end: datetime,
) -> None:
    """
    From `AEMO PASA Outputs <https://aemo.com.au/energy-systems/electricity/
    national-electricity-market-nem/data-nem/market-management-system-mms-data/
    projected-assessment-of-system-adequacy-pasa>`_:

      [MT PASA] is produced weekly (on Tuesdays) and lists the medium-term
      supply/demand prospects for the period two years in advance.
      The information is provided for each day within the report period.

    Noting that:

    - `MT PASA` is actually run at half-hourly resolution
      - But results are aggregated and reported for each day
    - Timing of "RUN_DATETIME" appears to be inconsistent on inspection
      - No validation on :attr:`run_start` and :attr:`run_end`
      - Compiler will instead collect all forecasts between provided forecast datetimes

    Validation checks:

    Check 1:
      `forecasted_start` and `forecasted_end` are at 00:00 for each supplied date.
      This is because results are reported for a day.
    Check 2:
      `forecasted_end` is within 2 years and 16 days of `run_end`. A 16 day offset
      appears to be in older data.Newer data appears to have a 6 day offset.

    Todo:
        Handle MTPASA DUID Availability

    Args:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
    Raises:
        ValueError: If any validation conditions are failed.
    """
    # Check 1
    for forecasted_input in (forecasted_start, forecasted_end):
        if forecasted_input.hour != 0 or forecasted_input.minute != 0:
            raise ValueError(
                "Results for MT PASA reported for each day. Forecasted start/end should"
                + " be supplied with hh:mm of 00:00"
            )
    # Check 2
    if run_end.month == 2 and run_end.day == 29:
        plus_two_years = run_end.replace(year=run_end.year + 2, day=28)
    else:
        plus_two_years = run_end.replace(year=run_end.year + 2)
    check_end_date = plus_two_years + timedelta(days=16)
    if forecasted_end > check_end_date:
        print_allowed = check_end_date.strftime(DATETIME_FORMAT)
        raise ValueError(
            f"For MT PASA, forecasted_end must be no later than {print_allowed} "
            + "based on the supplied run_end"
        )
    return None



from datetime import datetime, timedelta
from typing import Tuple

from ..data import DATETIME_FORMAT
from ..downloader import _validate_forecast_type
from ..query import _dt_converter
from .validators import (
    validate_MTPASA_datetime_inputs,
    validate_P5MIN_datetime_inputs,
    validate_PREDISPATCH_datetime_inputs,
    validate_STPASA_datetime_inputs,
)


def _determine_valid_earliest_run_for_PD(forecasted_dt: datetime) -> datetime:
    """Determine the earliest forecast run to use for a provided `forecasted` time

    For :term:`PREDISPATCH` and :term:`PDPASA` the 1300 forecast run will
    extend out to the last relevant interval of the trading day for which bid band price
    submission closed at 1230.

    Args:
        dt: A `forecasted` time.
    Returns:
        A 1300 run time, corresponding to the earliest possible run.

    """
    if forecasted_dt.hour > 4 or (forecasted_dt.hour == 4 and forecasted_dt.minute > 0):
        run_1300 = forecasted_dt - timedelta(days=1)
    else:
        run_1300 = forecasted_dt - timedelta(days=2)
    return run_1300.replace(hour=13, minute=0)


def _determine_valid_latest_run_for_STPASA(forecasted_dt: datetime) -> datetime:
    """Determines (one of the) latest forecast runs to use for a provided `forecasted`
    time

    :term:`STPASA` runs will produce forecasts that begin after the end of the
    `PREDISPATCH`/`PDPASA` forecast horizon.

    Prior to the change in the frequency at which ST PASA is run, 14:00 would be the
    latest forecast time. This has since changed to 13:00 (as ST PASA is now run
    hourly). To ensure "backwards-compatability", this function returns a run at 14:00.

    Args:
        dt: A `forecasted` time.
    Returns:
        (One of the) earliest possible run times.

    """
    if forecasted_dt.hour > 4 or (forecasted_dt.hour == 4 and forecasted_dt.minute > 0):
        run_1400 = forecasted_dt - timedelta(days=1)
    else:
        run_1400 = forecasted_dt - timedelta(days=2)
    return run_1400.replace(hour=14, minute=0)


def _generate_P5MIN_runtimes(
    forecasted_start: datetime, forecasted_end: datetime
) -> Tuple[datetime, datetime]:
    """Generates the earliest :term:`run_start` and latest :term:`run_end` for a set of
    user-supplied :term:`forecasted_start` and :term:`forecasted_end` times.

    Calls validation function to ensure that user-supplied `forecasted` times are valid.

    Args:
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
    Returns:
        Tuple of datetimes containing  the widest range of possible `forecasted` times
    """
    run_start = forecasted_start - timedelta(minutes=55)
    run_end = forecasted_end
    validate_P5MIN_datetime_inputs(run_start, run_end, forecasted_start, forecasted_end)
    return (run_start, run_end)


def _generate_PREDISPATCH_runtimes(
    forecasted_start: datetime, forecasted_end: datetime
) -> Tuple[datetime, datetime]:
    """Generates the earliest :term:`run_start` and latest :term:`run_end` for a set of
    user-supplied :term:`forecasted_start` and :term:`forecasted_end` times.

    Calls validation function to ensure that user-supplied `forecasted` times are valid.

    Args:
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
    Returns:
        Tuple of datetimes containing  the widest range of possible `forecasted` times
    """
    run_start = _determine_valid_earliest_run_for_PD(forecasted_start)
    run_end = forecasted_end
    validate_PREDISPATCH_datetime_inputs(
        run_start, run_end, forecasted_start, forecasted_end
    )
    return (run_start, run_end)


def _generate_PDPASA_runtimes(
    forecasted_start: datetime, forecasted_end: datetime
) -> Tuple[datetime, datetime]:
    """Generates the earliest :term:`run_start` and latest :term:`run_end` for a set of
    user-supplied :term:`forecasted_start` and :term:`forecasted_end` times.

    Calls validation function to ensure that user-supplied `forecasted` times are valid.

    Args:
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
    Returns:
        Tuple of datetimes containing  the widest range of possible `forecasted` times
    """

    (run_start, run_end) = _generate_PREDISPATCH_runtimes(
        forecasted_start, forecasted_end
    )
    return (run_start, run_end)


def _generate_STPASA_runtimes(
    forecasted_start: datetime, forecasted_end: datetime
) -> Tuple[datetime, datetime]:
    """Generates the earliest :term:`run_start` and latest :term:`run_end` for a set of
    user-supplied :term:`forecasted_start` and :term:`forecasted_end` times.

    Calls validation function to ensure that user-supplied `forecasted` times are valid.

    Args:
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
    Returns:
        Tuple of datetimes containing  the widest range of possible `forecasted` times
    """
    run_start = _determine_valid_latest_run_for_STPASA(forecasted_start) - timedelta(
        days=6
    )
    run_end = _determine_valid_latest_run_for_STPASA(forecasted_end)
    validate_STPASA_datetime_inputs(
        run_start, run_end, forecasted_start, forecasted_end
    )
    return (run_start, run_end)


def _generate_MTPASA_runtimes(
    forecasted_start: datetime, forecasted_end: datetime
) -> Tuple[datetime, datetime]:
    """Generates the earliest :term:`run_start` and latest :term:`run_end` for a set of
    user-supplied :term:`forecasted_start` and :term:`forecasted_end` times.

    Calls validation function to ensure that user-supplied `forecasted` times are valid.

    Args:
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
    Returns:
        Tuple of datetimes containing  the widest range of possible `forecasted` times
    """
    if forecasted_start.month == 2 and forecasted_start.day == 29:
        minus_two_years = forecasted_start.replace(
            year=forecasted_start.year - 2, day=28
        )
    else:
        minus_two_years = forecasted_start.replace(year=forecasted_start.year - 2)
    run_start = minus_two_years - timedelta(days=16)
    run_end = forecasted_end - timedelta(days=6)
    validate_MTPASA_datetime_inputs(
        run_start, run_end, forecasted_start, forecasted_end
    )
    return (run_start, run_end)


def generate_runtimes(
    forecasted_start: str, forecasted_end: str, forecast_type: str
) -> Tuple[str, str]:
    """For a particular :term:`forecast type`, generates valid `run` times provided that
    user-supplied `forecasted` times are valid.

    Args:
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retained.
        forecast_type: One of :data:`nemseer.forecast_types`
    Returns:
        Tuple of `nemseer`-valid string datetimes that correspond to valid `run` times
    Raises:
        ValueError: If supplied `forecasted` times are invalid.
    """
    _validate_forecast_type(forecast_type)
    if forecasted_start > forecasted_end:
        raise ValueError(
            "Forecasted end datetime must be greater than or equal to"
            + " forecasted start datetime."
        )
    generate_map = {
        "P5MIN": _generate_P5MIN_runtimes,
        "PREDISPATCH": _generate_PREDISPATCH_runtimes,
        "PDPASA": _generate_PDPASA_runtimes,
        "STPASA": _generate_STPASA_runtimes,
        "MTPASA": _generate_MTPASA_runtimes,
    }
    generate_func = generate_map[forecast_type]
    (run_start, run_end) = generate_func(
        _dt_converter(forecasted_start), _dt_converter(forecasted_end)
    )
    return (
        run_start.strftime(DATETIME_FORMAT),
        run_end.strftime(DATETIME_FORMAT),
    )



import logging
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Union

import pandas as pd
import xarray as xr
from attrs import define, field

from .data import ENUMERATED_TABLES, INVALID_STUBS_FILE
from .data_handlers import apply_run_and_forecasted_time_filters
from .forecast_type.validators import (
    validate_MTPASA_datetime_inputs,
    validate_P5MIN_datetime_inputs,
    validate_PDPASA_datetime_inputs,
    validate_PREDISPATCH_datetime_inputs,
    validate_STPASA_datetime_inputs,
)
from .query import Query, _enumerate_tables, generate_sqlloader_filenames

logger = logging.getLogger(__name__)


def _map_files_to_table(
    run_start: datetime,
    run_end: datetime,
    forecast_type: str,
    tables: List[str],
) -> Dict[str, List[str]]:
    """Maps filenames of interest to each queried table

    Translates output from `generate_sqlloader_filenames` to map filenames of interest
    to a queried table name.

    Args:
        forecast_start: Forecasts made at or after this datetime are queried.
        forecast_end: Forecasts made before or at this datetime are queried.
        forecast_type: One of :data:`nemseer.forecast_types`
        tables: Tables queried.
    Returns:
        A dictionary mapping the queried table name to filenames associated with that
        queried table.
    """
    metadata_to_filename = generate_sqlloader_filenames(
        run_start, run_end, forecast_type, tables
    )
    table_file_map = {}
    for table in tables:
        filenames_to_map = list()
        for metadata in metadata_to_filename.keys():
            if metadata[2] == table:
                filenames_to_map.append(metadata_to_filename[metadata])
        table_file_map[table] = filenames_to_map
    return table_file_map


def _input_datetime_validation(instance, attribute, value) -> None:
    """Dispatches the correct datetime validator based on the :term:`forecast_type`"""
    validator_map = {
        "P5MIN": validate_P5MIN_datetime_inputs,
        "PREDISPATCH": validate_PREDISPATCH_datetime_inputs,
        "PDPASA": validate_PDPASA_datetime_inputs,
        "STPASA": validate_STPASA_datetime_inputs,
        "MTPASA": validate_MTPASA_datetime_inputs,
    }
    validator_func = validator_map[instance.forecast_type]
    validator_func(
        instance.run_start,
        instance.run_end,
        instance.forecasted_start,
        instance.forecasted_end,
    )
    return None


@define
class DataCompiler:
    """:class:`DataCompiler` compiles data from the :term:`raw_cache` or
    :term:`processed_cache`.

    Attributes:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
        forecast_type: One of :data:`nemseer.forecast_types`.
        tables: Table or tables required. A single table can be supplied as
            a string. Multiple tables can be supplied as a list of strings.
        metadata: Metadata dictionary. Constructed by
            :meth:`Query.initialise() <nemseer.query.Query.initialise()>`.
        raw_cache (optional): Path to build or reuse :term:`raw_cache`.
        processed_cache (optional): Path to build or reuse :term`processed cache`.
            Should be distinct from :attr:`raw_cache`
        compiled_data: Defaults to `None` on initialisation. Populated once data
            is compiled by methods.
    """

    run_start: datetime = field(validator=_input_datetime_validation)
    run_end: datetime
    forecasted_start: datetime
    forecasted_end: datetime
    forecast_type: str
    tables: List[str]
    metadata: Dict
    raw_cache: Path
    processed_cache: Union[None, Path]
    compiled_data: Union[None, Dict[str, pd.DataFrame], Dict[str, xr.Dataset]] = field(
        default=None
    )

    @classmethod
    def from_Query(cls, query: Query) -> "DataCompiler":
        """Constructor method for :class:`DataCompiler` from
        :class:`Query <nemseer.query.Query>`."""
        tables = query.tables
        for ftype in ENUMERATED_TABLES:
            if query.forecast_type == ftype:
                for table, enumerate_to in ENUMERATED_TABLES[ftype]:
                    if table in tables:
                        tables = _enumerate_tables(tables, table, enumerate_to)
        if hasattr(query, "processed_cache"):
            processed_cache = query.processed_cache
        else:
            processed_cache = None
        return cls(
            run_start=query.run_start,
            run_end=query.run_end,
            forecasted_start=query.forecasted_start,
            forecasted_end=query.forecasted_end,
            forecast_type=query.forecast_type,
            tables=tables,
            metadata=query.metadata,
            raw_cache=query.raw_cache,
            processed_cache=processed_cache,
            compiled_data=None,
        )

    def invalid_or_corrupted_files(self) -> List[str]:
        """A list of invalid/corrupted files as per files in `.invalid_aemo_files.txt`.
        Returns an empty list if the stubfile does not exist.
        """
        invalid_or_corrupted_stubfile = self.raw_cache / Path(INVALID_STUBS_FILE)
        if invalid_or_corrupted_stubfile.exists():
            with open(invalid_or_corrupted_stubfile, "r") as f:
                invalid_or_corrupted = f.readlines()
            check_files = [f.strip() for f in invalid_or_corrupted]
            return check_files
        else:
            return []

    def compile_raw_data_to_dataframe(self) -> None:
        """Compiles data from :term:`raw_cache` to a :class:`pandas.DataFrame`

        This compiler will:

        - Skip invalid/corrupted files as recorded in `.invalid_aemo_files.txt`
        - Read :term:`raw_cache` parquet files and apply datetime filtering

        Returns:
            A single concatenated DataFrame if only one table type is requested.
            Otherwise, returns a dictionary with concatenated DataFrames mapped to each
            requested table type.
        Warning:
            Skips any files previously found to be invalid/corrupted and prints a
            warning
        """
        file_to_table_map = _map_files_to_table(
            self.run_start, self.run_end, self.forecast_type, self.tables
        )
        table_to_df_map = {}
        invalid_files = self.invalid_or_corrupted_files()
        for table in file_to_table_map.keys():
            files = file_to_table_map[table]
            filtered_files = [file for file in files if file not in invalid_files]
            if len(filtered_files) < len(files):
                logging.warning(
                    "Some files not compiled as these were found to be"
                    + " invalid/corrupt on previous download. You can force nemseer"
                    + " to load this file by deleting it from "
                    + ".invalid_aemo_files.txt"
                )
            dfs = []
            for file in filtered_files:
                filepath = self.raw_cache / Path(f"{file}.parquet")
                df = pd.read_parquet(filepath)
                df = apply_run_and_forecasted_time_filters(
                    df,
                    self.run_start,
                    self.run_end,
                    self.forecasted_start,
                    self.forecasted_end,
                    self.forecast_type,
                )
                dfs.append(df.reset_index(drop=True))
            if len(dfs) == 1:
                table_to_df_map[table] = dfs.pop()
            else:
                table_to_df_map[table] = pd.concat(dfs, axis=0)
        self.compiled_data = table_to_df_map



from typing import Dict, List, Union

import pandas as pd
import xarray as xr

from .data_compilers import DataCompiler
from .downloader import ForecastTypeDownloader
from .query import Query


def _initiate_downloads_from_query(query: Query, keep_csv: bool = False) -> None:
    """Initiates download actions using :class:`nemseer.query.Query`
    Args:
        query: :class:`nemseer.query.Query`
    Returns:
        None
    """
    if query.check_data_in_cache():
        pass
    else:
        downloader = ForecastTypeDownloader.from_Query(query)
        downloader.download_csv()
        downloader.convert_to_parquet(keep_csv=keep_csv)
    return None


def download_raw_data(
    run_start: str,
    run_end: str,
    forecasted_start: str,
    forecasted_end: str,
    forecast_type: str,
    tables: Union[str, List[str]],
    raw_cache: str,
    keep_csv: bool = False,
) -> None:
    """Downloads raw forecast data from NEMWeb MMSDM Historical Data SQLLoader

    Downloads raw forecast data and converts to parquet.

    Arguments:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retained.
        forecast_type: One of :data:`nemseer.forecast_types`
        tables: Table or tables required. A single table can be supplied as
            a string. Multiple tables can be supplied as a list of strings.
        raw_cache: Path to create or reuse as :term:`raw_cache`. Files are downloaded
            to this directory and cached data is maintained in the parquet format.
        keep_csv: Default False. If True, downloaded csvs are retained in the
            :term:`raw_cache`.
    """
    query = Query.initialise(
        run_start=run_start,
        run_end=run_end,
        forecasted_start=forecasted_start,
        forecasted_end=forecasted_end,
        forecast_type=forecast_type,
        tables=tables,
        raw_cache=raw_cache,
    )
    _initiate_downloads_from_query(query, keep_csv=keep_csv)


def compile_raw_data(
    run_start: str,
    run_end: str,
    forecasted_start: str,
    forecasted_end: str,
    forecast_type: str,
    tables: Union[str, List[str]],
    raw_cache: str,
    data_format: str = "df",
) -> Union[Dict[str, pd.DataFrame], Dict[str, xr.Dataset], None]:
    """Downloads raw forecast data from NEMWeb MMSDM Historical Data SQLLoader

    Downloads raw forecast data and converts to parquet.

    Arguments:
        run_start: Forecast runs at or after this datetime are queried.
        run_end: Forecast runs before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retained.
        forecast_type: One of :data:`nemseer.forecast_types`
        tables: Table or tables required. A single table can be supplied as
            a string. Multiple tables can be supplied as a list of strings.
        raw_cache: Path to create or reuse as :term:`raw_cache`. Files are downloaded
            to this directory and cached data is maintained in the parquet format.
        data_format: Default is 'df', which returns :class:`pandas DataFrame`.
            Can also request 'xr', which returns :class:`xarray.Dataset`.
    Todo:
        Implement xarray compiler
    """
    if data_format not in (fmts := ("df", "xr")):
        raise ValueError(f"Invalid data format. Formats include: {fmts}")
    query = Query.initialise(
        run_start=run_start,
        run_end=run_end,
        forecasted_start=forecasted_start,
        forecasted_end=forecasted_end,
        forecast_type=forecast_type,
        tables=tables,
        raw_cache=raw_cache,
    )
    _initiate_downloads_from_query(query, keep_csv=False)
    compiler = DataCompiler.from_Query(query)
    if data_format == "df":
        compiler.compile_raw_data_to_dataframe()
    data = compiler.compiled_data
    return data



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nemseer', 'nemseer.forecast_type']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21,<22',
 'beautifulsoup4>=4,<5',
 'netCDF4>=1.6.0,<2.0.0',
 'numpy>=1.23.0,<2.0.0',
 'packaging>=21.3,<22.0',
 'pandas>=1.2,<2.0',
 'psutil>=5.9.1,<6.0.0',
 'pyarrow>=9.0.0,<10.0.0',
 'requests>=2,<3',
 'tqdm>=4.64.0,<5.0.0',
 'xarray>=2022,<2023']

setup_kwargs = {
    'name': 'nemseer',
    'version': '0.5.0',
    'description': 'A package for downloading and handling forecasts for the National Electricity Market (NEM) from the Australian Energy Market Operator (AEMO).',
    'long_description': '# nemseer\n\n[![PyPI version](https://badge.fury.io/py/nemseer.svg)](https://badge.fury.io/py/nemseer)\n[![Continuous Integration and Deployment](https://github.com/UNSW-CEEM/NEMSEER/actions/workflows/cicd.yml/badge.svg)](https://github.com/UNSW-CEEM/NEMSEER/actions/workflows/cicd.yml)\n[![Documentation Status](https://readthedocs.org/projects/nemseer/badge/?version=latest)](https://nemseer.readthedocs.io/en/latest/?badge=latest)\n[![codecov](https://codecov.io/gh/UNSW-CEEM/NEMSEER/branch/master/graph/badge.svg?token=BO69YSQIGI)](https://codecov.io/gh/UNSW-CEEM/NEMSEER)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nA package for downloading and handling historical National Electricity Market (NEM) forecast data produced by the Australian Energy Market Operator (AEMO).\n\n## Installation\n\n```bash\npip install nemseer\n```\n\n## Overview\n\n`nemseer` allows you to access historical AEMO [pre-dispatch](https://aemo.com.au/en/energy-systems/electricity/national-electricity-market-nem/data-nem/market-management-system-mms-data/pre-dispatch) and [Projected Assessment of System Adequacy (PASA)](https://wa.aemo.com.au/energy-systems/electricity/national-electricity-market-nem/nem-forecasting-and-planning/forecasting-and-reliability/projected-assessment-of-system-adequacy) forecast[^1] data available through the [MMSDM Historical Data SQLLoader](https://nemseer.readthedocs.io/en/latest/glossary.html#term-MMSDM-Historical-Data-SQLLOader).\n\n![forecast_overview](docs/source/_static/forecast_timeframes.png)\n\n<sub><sup>Source: [Reserve services in the National Electricity Market, AEMC, 2021](https://www.aemc.gov.au/sites/default/files/2020-12/AEMC_Reserve%20services%20in%20the%20NEM%20directions%20paper_05.01.2021.pdf)</sup></sub>\n\nWhereas PASA processes are primarily used to assess resource adequacy based on technical inputs and assumptions for resources in the market (i.e. used to answer questions such as *"can operational demand be met in the forecast horizon with a sufficient safety (reserve) margin?"*), pre-dispatch processes incorporate the latest set of market participant offers and thus produce regional prices forecasts for energy and frequency control ancillary services [(FCAS)](https://aemo.com.au/-/media/files/electricity/nem/security_and_reliability/ancillary_services/guide-to-ancillary-services-in-the-national-electricity-market.pdf). Overviews of the various pre-dispatch and PASA processes can be found in the [glossary](https://nemseer.readthedocs.io/en/latest/glossary.html).\n\n[^1]: We use the term *"forecast"* loosely, especially given that these *"forecasts"* change once participants update offer information (e.g. through rebidding) or submit revised resource availabilities and energy constraints. Both of these are intended outcomes of these *"ahead processes"*, which are run to provide system and market information to participants to inform their decision-making. However, to avoid confusion and to ensure consistency with the language used by AEMO, we use the terms *"forecast"* (or outputs) and *"forecast types"* (or ahead processes) in `nemseer`.\n\n`nemseer` enables you to download and work with data from the following forecast types (where available, AEMO process and table descriptions are linked):\n\n1. 5-minute pre-dispatch (`P5MIN`: [Table descriptions](https://nemweb.com.au/Reports/Current/MMSDataModelReport/Electricity/MMS%20Data%20Model%20Report_files/MMS_222.htm#1))\n2. [Pre-dispatch](https://www.aemo.com.au/-/media/files/electricity/nem/security_and_reliability/power_system_ops/procedures/so_op_3704-predispatch.pdf?la=en) (`PREDISPATCH`: [Table descriptions](https://nemweb.com.au/Reports/Current/MMSDataModelReport/Electricity/MMS%20Data%20Model%20Report_files/MMS_260.htm#1))\n3. Pre-dispatch Projected Assessment of System Adequacy (`PDPASA`: [Tables and Descriptions](https://nemweb.com.au/Reports/Current/MMSDataModelReport/Electricity/MMS%20Data%20Model%20Report_files/MMS_467.htm#1))\n4. [Short Term Projected Assessment of System Adequacy](https://wa.aemo.com.au/-/media/files/electricity/nem/planning_and_forecasting/pasa/stpasa-process-description.pdf) (`STPASA`: [Table descriptions](https://nemweb.com.au/Reports/Current/MMSDataModelReport/Electricity/MMS%20Data%20Model%20Report_files/MMS_335.htm#1))\n5. [Medium Term Projected Assessment of System Adequacy](https://wa.aemo.com.au/-/media/files/electricity/nem/planning_and_forecasting/pasa/mt-pasa-process-description-v62.pdf?la=en) (`MTPASA`: [Table descriptions](https://nemweb.com.au/Reports/Current/MMSDataModelReport/Electricity/MMS%20Data%20Model%20Report_files/MMS_210.htm#1))\n\nAnother helpful reference for PASA information is AEMO\'s [Reliability Standard Implementation Guidelines](https://www.aemo.com.au/-/media/files/electricity/nem/planning_and_forecasting/rsig/reliability-standard-implementation-guidelines.pdf?la=en).\n\n### ST PASA Replacement Project\n\nNote that the methodologies for PD PASA and ST PASA are being reviewed by AEMO. In particular, the ST PASA Replacement project will combine PD PASA and ST PASA into ST PASA. For more detail, refer to the [final determination of the rule change](https://www.aemc.gov.au/sites/default/files/2022-05/ERC0332%20-%20Updating%20Short%20Term%20PASA%20-%20Final%20determination.pdf) and the [AEMO ST PASA Replacement Project home page](https://aemo.com.au/en/initiatives/trials-and-initiatives/st-pasa-replacement-project).\n\n## Usage\n\n### Glossary\n\nThe [glossary](https://nemseer.readthedocs.io/en/latest/glossary.html) contains overviews of the PASA and pre-dispatch processes, and descriptions of terminology used in `nemseer`.\n\n### Quick start\n\nCheck out the [Quick start](https://nemseer.readthedocs.io/en/latest/quick_start.html) for examples of how you can use `nemseer`.\n\n## Development progress\n\nWhile this package is currently under development, `nemseer` can currently be used to download, cache and handle historical forecast data from AEMO. Future development will focus on implementing advanced features, such as casting compiled data to [xarray](https://docs.xarray.dev/en/stable/index.html) n-dimensional arrays and caching processed queries in a netCDF format.\n\nFor a high-level overview of development, check out the [roadmap](./ROADMAP.md).\n\n## Contributing\n\nInterested in contributing? Check out the [contributing guidelines](./CONTRIBUTING.md), which also includes steps to install `nemseer` for development.\n\nPlease note that this project is released with a [Code of Conduct](./CONDUCT.md). By contributing to this project, you agree to abide by its terms.\n\n## License\n\n`nemseer` was created by Abhijith Prakash. It is licensed under the terms of the [BSD 3-Clause license](./LICENSE).\n\n## Credits\n\n`nemseer` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).\n\nThanks to Nicholas Gorman for his help in reviewing elements of `nemseer`.\n',
    'author': 'Abhijith Prakash',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
